# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.plugins.ml2 import api as driver_api
import six

BULK_EXTENDED = 'ml2plus:_bulk_extended'


@six.add_metaclass(abc.ABCMeta)
class SubnetPoolContext(object):
    """Context passed to MechanismDrivers for changes to subnet pool
    resources.

    A SubnetPoolContext instance wraps a subnet pool resource. It
    provides helper methods for accessing other relevant
    information. Results from expensive operations are cached so that
    other MechanismDrivers can freely access the same information.
    """

    @abc.abstractproperty
    def current(self):
        """Return the subnet pool in its current configuration.

        Return the subnet pool with all its properties 'current' at
        the time the context was established.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the subnet pool in its original configuration.

        Return the subnet pool, with all its properties set to their
        original values prior to a call to update_address_scope. Method is
        only valid within calls to update_address_scope_precommit and
        update_address_scope_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class AddressScopeContext(object):
    """Context passed to MechanismDrivers for changes to address scope
    resources.

    An AddressScopeContext instance wraps an address scope
    resource. It provides helper methods for accessing other relevant
    information. Results from expensive operations are cached so that
    other MechanismDrivers can freely access the same information.
    """

    @abc.abstractproperty
    def current(self):
        """Return the address scope in its current configuration.

        Return the address scope with all its properties 'current' at
        the time the context was established.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the address scope in its original configuration.

        Return the address scope, with all its properties set to their
        original values prior to a call to update_address_scope. Method is
        only valid within calls to update_address_scope_precommit and
        update_address_scope_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class SecurityGroupContext(object):
    """Context passed to MechanismDrivers for changes to security group
    resources.

    A SecurityGroupContext instance wraps a security group
    resource. It provides helper methods for accessing other relevant
    information. Results from expensive operations are cached so that
    other MechanismDrivers can freely access the same information.
    """

    @abc.abstractproperty
    def current(self):
        """Return the security group in its current configuration.

        Return the security group with all its properties 'current' at
        the time the context was established.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the security group in its original configuration.

        Return the security group, with all its properties set to their
        original values prior to a call to update_security_group. Method is
        only valid within calls to update_security_group_precommit and
        update_security_group_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class SecurityGroupRuleContext(object):
    """Context passed to MechanismDrivers for changes to security group
    rule resources.

    A SecurityGroupRuleContext instance wraps a security group rule
    resource. It provides helper methods for accessing other relevant
    information. Results from expensive operations are cached so that
    other MechanismDrivers can freely access the same information.
    """

    @abc.abstractproperty
    def current(self):
        """Return the security group rule in its current configuration.

        Return the security group rule with all its properties 'current' at
        the time the context was established.
        """
        pass

    @abc.abstractproperty
    def original(self):
        """Return the security group rule in its original configuration.

        Return the security group rule, with all its properties set to their
        original values prior to a call to update_security_group. Method is
        only valid within calls to update_security_group_rule_precommit and
        update_security_group_rule_postcommit.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class MechanismDriver(driver_api.MechanismDriver):

    def start_rpc_listeners(self):
        """Start driver-specify RPC listeners.

        If any driver-specific RPC listeners are needed, create an RPC
        connection, create the consumers, call consume_in_threads() on
        the connection, and return the resulting list of servers.
        """
        return []

    # REVISIT(rkukura): Is this needed for all operations, or just for
    # create operations? If its needed for all operations, should the
    # method be specific to the resource and operation, and include
    # the request data (i.e. update_network_pretransaction(self,
    # data))?
    def ensure_tenant(self, plugin_context, tenant_id):
        """Ensure tenant known before creating resource.

        :param plugin_context: Plugin request context.
        :param tenant_id: Tenant owning resource about to be created.

        Called before the start of a transaction creating any new core
        resource, allowing any needed tenant-specific processing to be
        performed.
        """
        pass

    def create_subnetpool_precommit(self, context):
        """Allocate resources for a new subnet pool.

        :param context: SubnetPoolContext instance describing the new
        subnet pool.

        Create a new subnet pool, allocating resources as necessary in
        the database. Called inside transaction context on
        session. Call cannot block.  Raising an exception will result
        in a rollback of the current transaction.
        """
        pass

    def create_subnetpool_postcommit(self, context):
        """Create a subnet pool.

        :param context: SubnetPoolContext instance describing the new
        subnet pool.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        """
        pass

    def update_subnetpool_precommit(self, context):
        """Update resources of a subnet pool.

        :param context: SubnetPoolContext instance describing the new
        state of the subnet pool, as well as the original state prior
        to the update_subnetpool call.

        Update values of a subnet pool, updating the associated
        resources in the database. Called inside transaction context
        on session.  Raising an exception will result in rollback of
        the transaction.

        update_subnetpool_precommit is called for all changes to the
        subnet pool state. It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        pass

    def update_subnetpool_postcommit(self, context):
        """Update a subnet pool.

        :param context: SubnetPoolContext instance describing the new
        state of the subnet pool, as well as the original state prior
        to the update_subnetpool call.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        update_subnetpool_postcommit is called for all changes to the
        subnet pool state.  It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.
        """
        pass

    def delete_subnetpool_precommit(self, context):
        """Delete resources for a subnet pool.

        :param context: SubnetPoolContext instance describing the
        current state of the subnet pool, prior to the call to delete
        it.

        Delete subnet pool resources previously allocated by this
        mechanism driver for a subnet pool. Called inside transaction
        context on session. Runtime errors are not expected, but
        raising an exception will result in rollback of the
        transaction.
        """
        pass

    def delete_subnetpool_postcommit(self, context):
        """Delete a subnet pool.

        :param context: SubnetPoolContext instance describing the
        current state of the subnet pool, prior to the call to delete
        it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        pass

    def create_address_scope_precommit(self, context):
        """Allocate resources for a new address scope.

        :param context: AddressScopeContext instance describing the
        new address scope.

        Create a new address scope, allocating resources as necessary
        in the database. Called inside transaction context on
        session. Call cannot block.  Raising an exception will result
        in a rollback of the current transaction.
        """
        pass

    def create_address_scope_postcommit(self, context):
        """Create an address scope.

        :param context: AddressScopeContext instance describing the
        new address scope.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        """
        pass

    def update_address_scope_precommit(self, context):
        """Update resources of an address scope.

        :param context: AddressScopeContext instance describing the
        new state of the address scope, as well as the original state
        prior to the update_address_scope call.

        Update values of an address scope, updating the associated
        resources in the database. Called inside transaction context
        on session.  Raising an exception will result in rollback of
        the transaction.

        update_address_scope_precommit is called for all changes to
        the address scope state. It is up to the mechanism driver to
        ignore state or state changes that it does not know or care
        about.
        """
        pass

    def update_address_scope_postcommit(self, context):
        """Update an address scope.

        :param context: AddressScopeContext instance describing the
        new state of the address scope, as well as the original state
        prior to the update_address_scope call.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        update_address_scope_postcommit is called for all changes to
        the address scope state.  It is up to the mechanism driver to
        ignore state or state changes that it does not know or care
        about.
        """
        pass

    def delete_address_scope_precommit(self, context):
        """Delete resources for an address scope.

        :param context: AddressScopeContext instance describing the
        current state of the address scope, prior to the call to
        delete it.

        Delete address scope resources previously allocated by this
        mechanism driver for an address scope. Called inside
        transaction context on session. Runtime errors are not
        expected, but raising an exception will result in rollback of
        the transaction.
        """
        pass

    def delete_address_scope_postcommit(self, context):
        """Delete an address scope.

        :param context: AddressScopeContext instance describing the
        current state of the address scope, prior to the call to
        delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        pass

    def create_security_group_precommit(self, context):
        """Allocate resources for a new security group.

        :param context: SecurityGroupContext instance describing the
        new security group.

        Create a new security group, allocating resources as necessary
        in the database. Called inside transaction context on
        session. Call cannot block.  Raising an exception will result
        in a rollback of the current transaction.
        """
        pass

    def create_security_group_postcommit(self, context):
        """Create a security group.

        :param context: SecurityGroupContext instance describing the
        new security group.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        This API is not being implemented at this moment.
        """
        pass

    def update_security_group_precommit(self, context):
        """Update resources of a security group.

        :param context: SecurityGroupContext instance describing the
        new state of the security group, as well as the original state
        prior to the update_security_group call.

        Update values of an security group, updating the associated
        resources in the database. Called inside transaction context
        on session.  Raising an exception will result in rollback of
        the transaction.

        update_security_group_precommit is called for all changes to
        the security group state. It is up to the mechanism driver to
        ignore state or state changes that it does not know or care
        about.
        """
        pass

    def update_security_group_postcommit(self, context):
        """Update a security group.

        :param context: SecurityGroupContext instance describing the
        new state of the security group, as well as the original state
        prior to the update_security_group call.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        update_security_group_postcommit is called for all changes to
        the security group state.  It is up to the mechanism driver to
        ignore state or state changes that it does not know or care
        about.

        This API is not being implemented at this moment.
        """
        pass

    def delete_security_group_precommit(self, context):
        """Delete resources for a security group.

        :param context: SecurityGroupContext instance describing the
        current state of the security group, prior to the call to
        delete it.

        Delete security group resources previously allocated by this
        mechanism driver for an security group. Called inside
        transaction context on session. Runtime errors are not
        expected, but raising an exception will result in rollback of
        the transaction.
        """
        pass

    def delete_security_group_postcommit(self, context):
        """Delete a security group.

        :param context: SecurityGroupContext instance describing the
        current state of the security group, prior to the call to
        delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.

        This API is not being implemented at this moment.
        """
        pass

    def create_security_group_rule_precommit(self, context):
        """Allocate resources for a new security group.

        :param context: SecurityGroupRuleContext instance describing the
        new security group rule.

        Create a new security group rule, allocating resources as necessary
        in the database. Called inside transaction context on
        session. Call cannot block.  Raising an exception will result
        in a rollback of the current transaction.
        """
        pass

    def create_security_group_rule_postcommit(self, context):
        """Create a security group rule.

        :param context: SecurityGroupRuleContext instance describing the
        new security group rule.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        This API is not being implemented at this moment.
        """
        pass

    # Security group rule updates are not supported by the Neutron API.

    def delete_security_group_rule_precommit(self, context):
        """Delete resources for a security group rule.

        :param context: SecurityGroupRuleContext instance describing the
        current state of the security group rule, prior to the call to
        delete it.

        Delete security group rule resources previously allocated by this
        mechanism driver for an security group rule. Called inside
        transaction context on session. Runtime errors are not
        expected, but raising an exception will result in rollback of
        the transaction.
        """
        pass

    def delete_security_group_rule_postcommit(self, context):
        """Delete a security group rule.

        :param context: SecurityGroupRuleContext instance describing the
        current state of the security group rule, prior to the call to
        delete it.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.

        This API is not being implemented at this moment.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class ExtensionDriver(driver_api.ExtensionDriver):

    def process_create_subnetpool(self, plugin_context, data, result):
        """Process extended attributes for create subnet pool.

        :param plugin_context: plugin request context
        :param data: dictionary of incoming subnet pool data
        :param result: subnet pool dictionary to extend

        Called inside transaction context on plugin_context.session to
        validate and persist any extended subnet pool attributes
        defined by this driver. Extended attribute values must also be
        added to result.
        """
        pass

    def process_update_subnetpool(self, plugin_context, data, result):
        """Process extended attributes for update subnet pool.

        :param plugin_context: plugin request context
        :param data: dictionary of incoming subnet pool data
        :param result: subnet pool dictionary to extend

        Called inside transaction context on plugin_context.session to
        validate and update any extended subnet pool attributes
        defined by this driver. Extended attribute values, whether
        updated or not, must also be added to result.
        """
        pass

    def extend_subnetpool_dict(self, session, base_model, result):
        """Add extended attributes to subnet pool dictionary.

        :param session: database session
        :param base_model: subnet pool model data
        :param result: subnet pool dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to a subnet pool
        dictionary to be used for mechanism driver calls and/or
        returned as the result of a subnet pool operation.
        """
        pass

    def process_create_address_scope(self, plugin_context, data, result):
        """Process extended attributes for create address scope.

        :param plugin_context: plugin request context
        :param data: dictionary of incoming address scope data
        :param result: address scope dictionary to extend

        Called inside transaction context on plugin_context.session to
        validate and persist any extended address scope attributes
        defined by this driver. Extended attribute values must also be
        added to result.
        """
        pass

    def process_update_address_scope(self, plugin_context, data, result):
        """Process extended attributes for update address scope.

        :param plugin_context: plugin request context
        :param data: dictionary of incoming address scope data
        :param result: address scope dictionary to extend

        Called inside transaction context on plugin_context.session to
        validate and update any extended address scope attributes
        defined by this driver. Extended attribute values, whether
        updated or not, must also be added to result.
        """
        pass

    def extend_address_scope_dict(self, session, base_model, result):
        """Add extended attributes to address scope dictionary.

        :param session: database session
        :param base_model: address scope model data
        :param result: address scope dictionary to extend

        Called inside transaction context on session to add any
        extended attributes defined by this driver to an address scope
        dictionary to be used for mechanism driver calls and/or
        returned as the result of an address scope operation.
        """
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nfp gateway db

Revision ID: 3e1f67cf951b
Revises: ce662ded3ba5
Create Date: 2016-11-28 04:07:43.928878

"""

# revision identifiers, used by Alembic.
revision = '3e1f67cf951b'
down_revision = 'ce662ded3ba5'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('nfp_service_gateway_info',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('network_function_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('gateway_ptg', sa.String(length=36),
                              nullable=False),
                    sa.Column('primary_instance_gw_pt', sa.String(length=36),
                              nullable=False),
                    sa.Column('secondary_instance_gw_pt', sa.String(length=36),
                              nullable=True),
                    sa.Column('primary_gw_vip_pt', sa.String(length=36),
                              nullable=True),
                    sa.Column('secondary_gw_vip_pt', sa.String(length=36),
                              nullable=True),
                    sa.ForeignKeyConstraint(['network_function_id'],
                                            ['nfp_network_functions.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id', 'network_function_id')
                    )

    op.add_column('nfp_network_function_devices', sa.Column('gateway_port',
                  sa.String(length=36), nullable=True))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""initial_qos_via_nsp

Revision ID: da6a25bbcfa8
Revises: cc09261e0fb5
Create Date: 2017-03-27 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = 'da6a25bbcfa8'
down_revision = 'cc09261e0fb5'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'gpm_qos_policy_mappings',
        sa.Column('service_policy_id', sa.String(length=36), nullable=False),
        sa.Column('qos_policy_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('service_policy_id'),
        sa.ForeignKeyConstraint(['service_policy_id'],
                                ['gp_network_service_policies.id'],
                                ondelete='CASCADE',
                                name='gbp_qos_policy_mapping_nsp_fk'),
        sa.ForeignKeyConstraint(['qos_policy_id'],
                                ['qos_policies.id'],
                                ondelete='RESTRICT',
                                name='gbp_qos_policy_mapping_qosp_fk')
    )


def downgrade():
    op.drop_table('gpm_qos_policy_mappings')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add l3out multiple epgs network extension

Revision ID: 8c5b556b4df1
Revises: e29a84f6a15f

"""

# revision identifiers, used by Alembic.
revision = '8c5b556b4df1'
down_revision = 'e29a84f6a15f'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql


def upgrade():
    op.add_column('apic_aim_network_extensions',
                  sa.Column('multi_ext_nets', sa.Boolean,
                            nullable=False, server_default=sql.false()))
    pass


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from unittest import mock

from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.extensions import base as test_extensions_base
from neutron_lib.plugins import constants
from oslo_utils import uuidutils
from webob import exc

from gbpservice.neutron.extensions import servicechain
from gbpservice.neutron.tests.unit import common as cm


_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path
SERVICE_CHAIN_PLUGIN_BASE_NAME = (
    servicechain.ServiceChainPluginBase.__module__ + '.' +
    servicechain.ServiceChainPluginBase.__name__)
SERVICECHAIN_URI = 'servicechain'
SERVICECHAIN_NODES_URI = SERVICECHAIN_URI + '/' + 'servicechain_nodes'
SERVICECHAIN_SPECS_URI = SERVICECHAIN_URI + '/' + 'servicechain_specs'
SERVICECHAIN_INSTANCES_URI = SERVICECHAIN_URI + '/' + 'servicechain_instances'
SERVICE_PROFILE_URI = SERVICECHAIN_URI + '/' + 'service_profiles'


class ServiceChainExtensionTestCase(test_extensions_base.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        super(ServiceChainExtensionTestCase, self).setUp()
        plural_mappings = {}
        self.setup_extension(
            SERVICE_CHAIN_PLUGIN_BASE_NAME, constants.SERVICECHAIN,
            servicechain.Servicechain,
            SERVICECHAIN_URI, plural_mappings=plural_mappings)
        self.instance = self.plugin.return_value

    def _test_create_servicechain_node(self, data, expected_value,
                                       default_data=None):
        if not default_data:
            default_data = data
        self.instance.create_servicechain_node.return_value = expected_value
        res = self.api.post(_get_path(SERVICECHAIN_NODES_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)

        self.instance.create_servicechain_node.assert_called_once_with(
            mock.ANY, servicechain_node=default_data)

        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_node', res)
        self.assertEqual(expected_value, res['servicechain_node'])

    def test_create_servicechain_node_with_defaults(self):
        servicechain_node_id = _uuid()
        data = {
            'servicechain_node': {
                'service_profile_id': _uuid(),
                'tenant_id': _uuid(),
                'config': 'test_config',
                'service_type': None,
            }
        }
        default_attrs = (
            cm.get_create_servicechain_node_default_attrs_and_prj_id())
        default_data = copy.copy(data)
        default_data['servicechain_node'].update(default_attrs)
        expected_value = dict(default_data['servicechain_node'])
        expected_value['id'] = servicechain_node_id

        self._test_create_servicechain_node(data, expected_value, default_data)

    def test_create_servicechain_node(self):
        servicechain_node_id = _uuid()
        data = {
            'servicechain_node': cm.get_create_servicechain_node_attrs()
        }
        expected_value = dict(data['servicechain_node'])
        expected_value['id'] = servicechain_node_id

        self._test_create_servicechain_node(data, expected_value)

    def test_list_servicechain_nodes(self):
        servicechain_node_id = _uuid()
        expected_value = [{'tenant_id': _uuid(), 'id': servicechain_node_id}]
        self.instance.get_servicechain_nodes.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_NODES_URI, fmt=self.fmt))

        self.instance.get_servicechain_nodes.assert_called_once_with(
            mock.ANY, fields=mock.ANY, filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_nodes', res)
        self.assertEqual(expected_value, res['servicechain_nodes'])

    def test_get_servicechain_node(self):
        servicechain_node_id = _uuid()
        expected_value = {
            'tenant_id': _uuid(), 'id': servicechain_node_id}
        self.instance.get_servicechain_node.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_NODES_URI,
                                     id=servicechain_node_id,
                                     fmt=self.fmt))

        self.instance.get_servicechain_node.assert_called_once_with(
            mock.ANY, servicechain_node_id, fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_node', res)
        self.assertEqual(expected_value, res['servicechain_node'])

    def test_update_servicechain_node(self):
        servicechain_node_id = _uuid()
        update_data = {
            'servicechain_node': cm.get_update_servicechain_node_attrs()
        }
        expected_value = {'tenant_id': _uuid(), 'id': servicechain_node_id}
        self.instance.update_servicechain_node.return_value = expected_value

        res = self.api.put(_get_path(SERVICECHAIN_NODES_URI,
                                     id=servicechain_node_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        self.instance.update_servicechain_node.assert_called_once_with(
            mock.ANY, servicechain_node_id, servicechain_node=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_node', res)
        self.assertEqual(expected_value, res['servicechain_node'])

    def test_delete_servicechain_node(self):
        self._test_entity_delete('servicechain_node')

    def _test_create_servicechain_spec(self, data, expected_value,
                                       default_data=None):
        if not default_data:
            default_data = data
        self.instance.create_servicechain_spec.return_value = expected_value

        res = self.api.post(_get_path(SERVICECHAIN_SPECS_URI, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)

        self.instance.create_servicechain_spec.assert_called_once_with(
            mock.ANY, servicechain_spec=default_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual(expected_value, res['servicechain_spec'])

    def test_create_servicechain_spec_with_defaults(self):
        servicechain_spec_id = _uuid()
        data = {
            'servicechain_spec': {
                'nodes': [_uuid(), _uuid()], 'tenant_id': _uuid()
            }
        }
        default_attrs = (
            cm.get_create_servicechain_spec_default_attrs_and_prj_id())
        default_data = copy.copy(data)
        default_data['servicechain_spec'].update(default_attrs)
        expected_value = dict(default_data['servicechain_spec'])
        expected_value['id'] = servicechain_spec_id

        self._test_create_servicechain_spec(data, expected_value, default_data)

    def test_create_servicechain_spec(self):
        servicechain_spec_id = _uuid()
        data = {
            'servicechain_spec': cm.get_create_servicechain_spec_attrs()
        }
        expected_value = dict(data['servicechain_spec'])
        expected_value['id'] = servicechain_spec_id

        self._test_create_servicechain_spec(data, expected_value)

    def test_list_servicechain_specs(self):
        servicechain_spec_id = _uuid()
        expected_value = [{'tenant_id': _uuid(), 'id': servicechain_spec_id}]
        self.instance.get_servicechain_specs.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_SPECS_URI, fmt=self.fmt))

        self.instance.get_servicechain_specs.assert_called_once_with(
            mock.ANY, fields=mock.ANY, filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_specs', res)
        self.assertEqual(expected_value, res['servicechain_specs'])

    def test_get_servicechain_spec(self):
        servicechain_spec_id = _uuid()
        expected_value = {'tenant_id': _uuid(), 'id': servicechain_spec_id}
        self.instance.get_servicechain_spec.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_SPECS_URI,
                                     id=servicechain_spec_id,
                                     fmt=self.fmt))

        self.instance.get_servicechain_spec.assert_called_once_with(
            mock.ANY, servicechain_spec_id, fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual(expected_value, res['servicechain_spec'])

    def test_update_servicechain_spec(self):
        servicechain_spec_id = _uuid()
        update_data = {
            'servicechain_spec': cm.get_update_servicechain_spec_attrs()
        }
        expected_value = {'tenant_id': _uuid(), 'id': servicechain_spec_id}
        self.instance.update_servicechain_spec.return_value = expected_value

        res = self.api.put(_get_path(SERVICECHAIN_SPECS_URI,
                                     id=servicechain_spec_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        self.instance.update_servicechain_spec.assert_called_once_with(
            mock.ANY, servicechain_spec_id, servicechain_spec=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual(expected_value, res['servicechain_spec'])

    def test_delete_servicechain_spec(self):
        self._test_entity_delete('servicechain_spec')

    def _test_create_servicechain_instance(self, data, expected_value,
                                           default_data=None):
        if not default_data:
            default_data = data
        self.instance.create_servicechain_instance.return_value = (
            expected_value)
        res = self.api.post(_get_path(SERVICECHAIN_INSTANCES_URI,
                                      fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)

        self.instance.create_servicechain_instance.assert_called_once_with(
            mock.ANY, servicechain_instance=default_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_instance', res)
        self.assertEqual(expected_value, res['servicechain_instance'])

    def test_create_servicechain_instance_with_defaults(self):
        servicechain_instance_id = _uuid()
        data = {
            'servicechain_instance': {
                'servicechain_specs': [_uuid()],
                'tenant_id': _uuid(),
                'provider_ptg_id': _uuid(),
                'consumer_ptg_id': _uuid(),
                'management_ptg_id': _uuid(),
                'classifier_id': _uuid(),
            }
        }
        default_attrs = (
            cm.get_create_servicechain_instance_default_attrs_and_prj_id())
        default_data = copy.copy(data)
        default_data['servicechain_instance'].update(default_attrs)
        expected_value = dict(default_data['servicechain_instance'])
        expected_value['id'] = servicechain_instance_id

        self._test_create_servicechain_instance(data, expected_value,
                                                default_data)

    def test_create_servicechain_instance(self):
        servicechain_instance_id = _uuid()
        data = {'servicechain_instance':
                cm.get_create_servicechain_instance_attrs()}
        expected_value = dict(data['servicechain_instance'])
        expected_value['id'] = servicechain_instance_id

        self._test_create_servicechain_instance(data, expected_value)

    def test_list_servicechain_instances(self):
        servicechain_instance_id = _uuid()
        expected_value = [{'tenant_id': _uuid(),
                           'id': servicechain_instance_id}]
        self.instance.get_servicechain_instances.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_INSTANCES_URI, fmt=self.fmt))

        self.instance.get_servicechain_instances.assert_called_once_with(
            mock.ANY, fields=mock.ANY, filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_instances', res)
        self.assertEqual(expected_value, res['servicechain_instances'])

    def test_get_servicechain_instance(self):
        servicechain_instance_id = _uuid()
        expected_value = {'tenant_id': _uuid(), 'id': servicechain_instance_id}
        self.instance.get_servicechain_instance.return_value = expected_value

        res = self.api.get(_get_path(SERVICECHAIN_INSTANCES_URI,
                                     id=servicechain_instance_id,
                                     fmt=self.fmt))

        self.instance.get_servicechain_instance.assert_called_once_with(
            mock.ANY, servicechain_instance_id, fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_instance', res)
        self.assertEqual(expected_value, res['servicechain_instance'])

    def test_update_servicechain_instance(self):
        servicechain_instance_id = _uuid()
        update_data = {'servicechain_instance':
                       cm.get_update_servicechain_instance_attrs()}
        expected_value = {'tenant_id': _uuid(), 'id': servicechain_instance_id}
        self.instance.update_servicechain_instance.return_value = (
            expected_value)

        res = self.api.put(_get_path(SERVICECHAIN_INSTANCES_URI,
                                     id=servicechain_instance_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        self.instance.update_servicechain_instance.assert_called_once_with(
            mock.ANY, servicechain_instance_id,
            servicechain_instance=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('servicechain_instance', res)
        self.assertEqual(expected_value, res['servicechain_instance'])

    def test_delete_servicechain_instance(self):
        self._test_entity_delete('servicechain_instance')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add snat subnet only extension for subnets

Revision ID: f8dc1eb9deaf
Revises: 872bf4ba86a6

"""

# revision identifiers, used by Alembic.
revision = 'f8dc1eb9deaf'
down_revision = '872bf4ba86a6'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('apic_aim_subnet_extensions',
                  sa.Column('snat_subnet_only',
                            sa.Boolean))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa

from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.tests.unit.services.grouppolicy.extensions import (
    test_extension as test_extension)


class TestPolicyTargetExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_target_extension'
    policy_target_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_policy_targets.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    pt_extension = sa.Column(sa.String(64))


class TestPolicyTargetGroupExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_target_group_extension'
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id',
                                     ondelete="CASCADE"),
        primary_key=True)
    ptg_extension = sa.Column(sa.String(64))


class TestL2PolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_l2_policy_extension'
    l2_policy_id = sa.Column(sa.String(36), sa.ForeignKey('gp_l2_policies.id',
                                                          ondelete="CASCADE"),
                             primary_key=True)
    l2p_extension = sa.Column(sa.String(64))


class TestL3PolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_l3_policy_extension'
    l3_policy_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_l3_policies.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    l3p_extension = sa.Column(sa.String(64))


class TestPolicyClassifierExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_classifier_extension'
    policy_classifier_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_classifiers.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    pc_extension = sa.Column(sa.String(64))


class TestPolicyActionExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_action_extension'
    policy_action_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_actions.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    pa_extension = sa.Column(sa.String(64))


class TestPolicyRuleExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_rule_extension'
    policy_rule_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_policy_rules.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    pr_extension = sa.Column(sa.String(64))


class TestPolicyRuleSetExtension(model_base.BASEV2):
    __tablename__ = 'test_policy_rule_set_extension'
    policy_rule_set_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_policy_rule_sets.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    prs_extension = sa.Column(sa.String(64))


class TestNetworkServicePolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_network_service_policy_extension'
    network_service_policy_id = sa.Column(sa.String(36),
                       sa.ForeignKey('gp_network_service_policies.id',
                                     ondelete="CASCADE"),
                       primary_key=True)
    nsp_extension = sa.Column(sa.String(64))


class TestExternalSegmentExtension(model_base.BASEV2):
    __tablename__ = 'test_external_segment_extension'
    external_segment_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_external_segments.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    es_extension = sa.Column(sa.String(64))


class TestExternalPolicyExtension(model_base.BASEV2):
    __tablename__ = 'test_external_policy_extension'
    external_policy_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_external_policies.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    ep_extension = sa.Column(sa.String(64))


class TestNatPoolExtension(model_base.BASEV2):
    __tablename__ = 'test_nat_pool_extension'
    nat_pool_id = sa.Column(sa.String(36),
                      sa.ForeignKey('gp_nat_pools.id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    np_extension = sa.Column(sa.String(64))


class TestExtensionDriver(api.ExtensionDriver):
    _supported_extension_alias = 'test_extension'
    _extension_dict = test_extension.EXTENDED_ATTRIBUTES_2_0

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    @api.default_extension_behavior(TestPolicyTargetExtension)
    def process_create_policy_target(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyTargetExtension)
    def process_update_policy_target(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyTargetExtension)
    def extend_policy_target_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestPolicyTargetGroupExtension)
    def process_create_policy_target_group(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyTargetGroupExtension)
    def process_update_policy_target_group(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyTargetGroupExtension)
    def extend_policy_target_group_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestL2PolicyExtension)
    def process_create_l2_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestL2PolicyExtension)
    def process_update_l2_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestL2PolicyExtension)
    def extend_l2_policy_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestL3PolicyExtension)
    def process_create_l3_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestL3PolicyExtension)
    def process_update_l3_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestL3PolicyExtension)
    def extend_l3_policy_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestPolicyClassifierExtension)
    def process_create_policy_classifier(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyClassifierExtension)
    def process_update_policy_classifier(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyClassifierExtension)
    def extend_policy_classifier_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestPolicyActionExtension)
    def process_create_policy_action(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyActionExtension)
    def process_update_policy_action(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyActionExtension)
    def extend_policy_action_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleExtension)
    def process_create_policy_rule(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleExtension)
    def process_update_policy_rule(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleExtension)
    def extend_policy_rule_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleSetExtension)
    def process_create_policy_rule_set(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleSetExtension)
    def process_update_policy_rule_set(self, session, data, result):
        pass

    @api.default_extension_behavior(TestPolicyRuleSetExtension)
    def extend_policy_rule_set_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestNetworkServicePolicyExtension)
    def process_create_network_service_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestNetworkServicePolicyExtension)
    def process_update_network_service_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestNetworkServicePolicyExtension)
    def extend_network_service_policy_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestExternalSegmentExtension)
    def process_create_external_segment(self, session, data, result):
        pass

    @api.default_extension_behavior(TestExternalSegmentExtension)
    def process_update_external_segment(self, session, data, result):
        pass

    @api.default_extension_behavior(TestExternalSegmentExtension)
    def extend_external_segment_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestExternalPolicyExtension)
    def process_create_external_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestExternalPolicyExtension)
    def process_update_external_policy(self, session, data, result):
        pass

    @api.default_extension_behavior(TestExternalPolicyExtension)
    def extend_external_policy_dict(self, session, result):
        pass

    @api.default_extension_behavior(TestNatPoolExtension)
    def process_create_nat_pool(self, session, data, result):
        pass

    @api.default_extension_behavior(TestNatPoolExtension)
    def process_update_nat_pool(self, session, data, result):
        pass

    @api.default_extension_behavior(TestNatPoolExtension)
    def extend_nat_pool_dict(self, session, result):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add policy enforcement pref attribute

Revision ID: 872bf4ba86a6
Revises: 016a678fafd4

"""

# revision identifiers, used by Alembic.
revision = '872bf4ba86a6'
down_revision = '016a678fafd4'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('apic_aim_network_extensions',
                  sa.Column('policy_enforcement_pref',
                            sa.Enum('unenforced', 'enforced', ''),
                            server_default="unenforced",
                            nullable=False))


def downgrade():
    pass



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import functools

from neutron_lib.api import converters as conv
from neutron_lib.api.definitions import address_scope as as_def
from neutron_lib.api.definitions import network as net_def
from neutron_lib.api.definitions import port as port_def
from neutron_lib.api.definitions import subnet as subnet_def
from neutron_lib.api import extensions
from neutron_lib.api import validators as valid
from oslo_log import log as logging
import six

from gbpservice._i18n import _

ALIAS = 'cisco-apic'

DIST_NAMES = 'apic:distinguished_names'
SYNC_STATE = 'apic:synchronization_state'
NAT_TYPE = 'apic:nat_type'
SNAT_HOST_POOL = 'apic:snat_host_pool'
ACTIVE_ACTIVE_AAP = 'apic:active_active_aap'
EXTERNAL_CIDRS = 'apic:external_cidrs'
SVI = 'apic:svi'
BGP = 'apic:bgp_enable'
BGP_ASN = 'apic:bgp_asn'
BGP_TYPE = 'apic:bgp_type'
NESTED_DOMAIN_NAME = 'apic:nested_domain_name'
NESTED_DOMAIN_TYPE = 'apic:nested_domain_type'
NESTED_DOMAIN_INFRA_VLAN = 'apic:nested_domain_infra_vlan'
NESTED_DOMAIN_ALLOWED_VLANS = 'apic:nested_domain_allowed_vlans'
NESTED_DOMAIN_SERVICE_VLAN = 'apic:nested_domain_service_vlan'
NESTED_DOMAIN_NODE_NETWORK_VLAN = 'apic:nested_domain_node_network_vlan'
EXTRA_PROVIDED_CONTRACTS = 'apic:extra_provided_contracts'
EXTRA_CONSUMED_CONTRACTS = 'apic:extra_consumed_contracts'
EPG_CONTRACT_MASTERS = 'apic:epg_contract_masters'
ERSPAN_CONFIG = 'apic:erspan_config'
POLICY_ENFORCEMENT_PREF = 'apic:policy_enforcement_pref'
SNAT_SUBNET_ONLY = 'apic:snat_subnet_only'
EPG_SUBNET = 'apic:epg_subnet'
NO_NAT_CIDRS = 'apic:no_nat_cidrs'
MULTI_EXT_NETS = 'apic:multi_ext_nets'

BD = 'BridgeDomain'
EPG = 'EndpointGroup'
SUBNET = 'Subnet'
VRF = 'VRF'
EXTERNAL_NETWORK = 'ExternalNetwork'
AP = 'ApplicationProfile'

SYNC_SYNCED = 'synced'
SYNC_BUILD = 'build'
SYNC_ERROR = 'error'
SYNC_NOT_APPLICABLE = 'N/A'

VLANS_LIST = 'vlans_list'
VLAN_RANGES = 'vlan_ranges'
APIC_MAX_VLAN = 4093
APIC_MIN_VLAN = 1
VLAN_RANGE_START = 'start'
VLAN_RANGE_END = 'end'

ERSPAN_DEST_IP = 'dest_ip'
ERSPAN_FLOW_ID = 'flow_id'
ERSPAN_DIRECTION = 'direction'

LOG = logging.getLogger(__name__)


def _validate_apic_vlan(data, key_specs=None):
    if data is None:
        return
    try:
        val = int(data)
        if val >= APIC_MIN_VLAN and val <= APIC_MAX_VLAN:
            return
        msg = ("Invalid value for VLAN: '%s'") % data
        LOG.debug(msg)
        return msg
    except (ValueError, TypeError):
        msg = ("Invalid data format for VLAN: '%s'") % data
        LOG.debug(msg)
        return msg


def _validate_apic_vlan_range(data, key_specs=None):
    if data is None:
        return

    expected_keys = [VLAN_RANGE_START, VLAN_RANGE_END]
    msg = valid._verify_dict_keys(expected_keys, data)
    if msg:
        return msg
    for k in expected_keys:
        msg = _validate_apic_vlan(data[k])
        if msg:
            return msg
    if int(data[VLAN_RANGE_START]) > int(data[VLAN_RANGE_END]):
        msg = ("Invalid start, end for VLAN range %s") % data
        return msg


def _validate_erspan_flow_id(data, key_specs=None):
    if data is None:
        return
    msg = valid.validate_non_negative(data)
    if int(data) > 1023:
        msg = ("ERSPAN flow ID must be less than 1023 (was %s)") % data
    elif int(data) == 0:
        msg = ("ERSPAN flow ID must be greater than 0 (was %s)") % data
    return msg


def _validate_erspan_configs(data, valid_values=None):
    """Validate a list of unique ERSPAN configurations.

    :param data: The data to validate. To be valid it must be a list like
        structure of ERSPAN config dicts, each containing 'dest_ip' and
        'flow_id' key values.
    :param valid_values: Not used!
    :returns: None if data is a valid list of unique ERSPAN config dicts,
        otherwise a human readable message indicating why validation failed.
    """
    if not isinstance(data, list):
        msg = ("Invalid data format for ERSPAN config: '%s'") % data
        LOG.debug(msg)
        return msg

    expected_keys = (ERSPAN_DEST_IP, ERSPAN_FLOW_ID,)
    erspan_configs = []
    for erspan_config in data:
        msg = valid._verify_dict_keys(expected_keys, erspan_config, False)
        if msg:
            return msg
        msg = _validate_erspan_flow_id(erspan_config[ERSPAN_FLOW_ID])
        if msg:
            return msg
        msg = valid.validate_ip_address(erspan_config[ERSPAN_DEST_IP])
        if msg:
            return msg
        if erspan_config in erspan_configs:
            msg = ("Duplicate ERSPAN config '%s'") % erspan_config
            LOG.debug(msg)
            return msg
        erspan_configs.append(erspan_config)


def _validate_dict_or_string(data, key_specs=None):
    if data is None:
        return

    if isinstance(data, str) or isinstance(data, six.string_types):
        try:
            data = ast.literal_eval(data)
        except Exception:
            msg = _("Extension %s cannot be converted to dict") % data
            return msg

    return valid.validate_dict_or_none(data, key_specs)


def convert_apic_vlan(value):
    if value is None:
        return
    else:
        return int(value)


def convert_apic_none_to_empty_list(value):
    if value is None:
        return []
    if isinstance(value, str) or isinstance(value, six.string_types):
        value = ast.literal_eval(value)
    return value


def convert_nested_domain_allowed_vlans(value):
    if value is None:
        return

    if isinstance(value, str) or isinstance(value, six.string_types):
        value = ast.literal_eval(value)

    vlans_list = []
    if VLANS_LIST in value:
        for vlan in value[VLANS_LIST]:
            vlans_list.append(convert_apic_vlan(vlan))
    if VLAN_RANGES in value:
        for vlan_range in value[VLAN_RANGES]:
            for vrng in [VLAN_RANGE_START, VLAN_RANGE_END]:
                vlan_range[vrng] = convert_apic_vlan(vlan_range[vrng])
            vlans_list.extend(list(range(vlan_range[VLAN_RANGE_START],
                vlan_range[VLAN_RANGE_END] + 1)))
    # eliminate duplicates
    vlans_list = list(set(vlans_list))
    # sort
    vlans_list.sort()
    value[VLANS_LIST] = vlans_list
    return value


valid.validators['type:apic_vlan'] = _validate_apic_vlan
valid.validators['type:apic_vlan_list'] = functools.partial(
        valid._validate_list_of_items, _validate_apic_vlan)
valid.validators['type:apic_vlan_range_list'] = functools.partial(
        valid._validate_list_of_items, _validate_apic_vlan_range)
valid.validators['type:dict_or_string'] = _validate_dict_or_string
valid.validators['type:apic_erspan_flow_id'] = _validate_erspan_flow_id
valid.validators['type:apic_erspan_configs'] = _validate_erspan_configs


APIC_ATTRIBUTES = {
    DIST_NAMES: {'allow_post': False, 'allow_put': False, 'is_visible': True},
    SYNC_STATE: {'allow_post': False, 'allow_put': False, 'is_visible': True}
}

ERSPAN_KEY_SPECS = [
    {ERSPAN_DEST_IP: {'type:ip_address': None,
                      'required': True},
     ERSPAN_FLOW_ID: {'type:apic_erspan_flow_id': None,
                      'required': True},
     ERSPAN_DIRECTION: {'type:values': ['in', 'out', 'both'],
                        'default': 'both'}},
]

EPG_CONTRACT_MASTER_KEY_SPECS = [
    # key spec for opt_name in _VALID_BLANK_EXTRA_DHCP_OPTS
    {'app_profile_name': {'type:not_empty_string': None,
                          'required': True},
     'name': {'type:not_empty_string': None,
              'required': True}},
]

PORT_ATTRIBUTES = {
    ERSPAN_CONFIG: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:apic_erspan_configs': None},
    },
}

NET_ATTRIBUTES = {
    SVI: {
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    BGP: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    BGP_TYPE: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': 'default_export',
        'validate': {'type:values': ['default_export', '']},
    },
    BGP_ASN: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': "0",
        'validate': {'type:non_negative': None},
    },
    NESTED_DOMAIN_NAME: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': '',
        'validate': {'type:string': None},
    },
    NESTED_DOMAIN_TYPE: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': '',
        'validate': {'type:string': None},
    },
    NESTED_DOMAIN_INFRA_VLAN: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'validate': {'type:apic_vlan': None},
        'convert_to': convert_apic_vlan,
    },
    NESTED_DOMAIN_SERVICE_VLAN: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'validate': {'type:apic_vlan': None},
        'convert_to': convert_apic_vlan,
    },
    NESTED_DOMAIN_NODE_NETWORK_VLAN: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'validate': {'type:apic_vlan': None},
        'convert_to': convert_apic_vlan,
    },
    NESTED_DOMAIN_ALLOWED_VLANS: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'validate': {
            'type:dict_or_string': {
                VLANS_LIST: {'type:apic_vlan_list': None},
                VLAN_RANGES: {'type:apic_vlan_range_list': None},
            }
        },
        'convert_to': convert_nested_domain_allowed_vlans,
    },
    EXTRA_PROVIDED_CONTRACTS: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:list_of_unique_strings': None},
    },
    EXTRA_CONSUMED_CONTRACTS: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:list_of_unique_strings': None},
    },
    EPG_CONTRACT_MASTERS: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:list_of_any_key_specs_or_none':
                     EPG_CONTRACT_MASTER_KEY_SPECS},
    },
    DIST_NAMES: {
        # DN of corresponding APIC L3Out external network or BD.
        # It can be specified only on create.
        # Change 'allow_put' if updates on other DNs is allowed later,
        # and validate that ExternalNetwork DN may not be updated.
        'allow_post': True, 'allow_put': False,
        'is_visible': True,
        'default': None,
        'validate': {
            'type:dict_or_none': {
                EXTERNAL_NETWORK: {'type:string': None,
                                   'required': False},
                BD: {'type:string': None,
                     'required': False}
            },
        }
    },
    POLICY_ENFORCEMENT_PREF: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': 'unenforced',
        'validate': {'type:values': ['unenforced', 'enforced', '']},
    },
    NO_NAT_CIDRS: {
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': None,
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:list_of_unique_strings': None},
    },
    MULTI_EXT_NETS: {
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
}

EXT_NET_ATTRIBUTES = {
    NAT_TYPE: {
        # whether NAT is enabled, and if so its type
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': 'distributed',
        'validate': {'type:values': ['distributed', 'edge', '']},
    },
    EXTERNAL_CIDRS: {
        # Restrict external traffic to specified addresses
        'allow_put': True, 'allow_post': True,
        'is_visible': True, 'default': ['0.0.0.0/0'],
        'convert_to': convert_apic_none_to_empty_list,
        'validate': {'type:subnet_list': None},
    },
}

EXT_SUBNET_ATTRIBUTES = {
    SNAT_HOST_POOL: {
        # Whether an external subnet should be used as a pool
        # for allocating host-based SNAT addresses.
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    ACTIVE_ACTIVE_AAP: {
        # Whether a subnet will support the active active AAP or not.
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    SNAT_SUBNET_ONLY: {
        # Whether this subnet can be used for assigning snat addresses only
        'allow_post': True, 'allow_put': True,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    },
    EPG_SUBNET: {
        # Whether this subnet is EPG subnet or regular subnet
        'allow_post': True, 'allow_put': False,
        'is_visible': True, 'default': False,
        'convert_to': conv.convert_to_boolean,
    }
}

ADDRESS_SCOPE_ATTRIBUTES = {
    DIST_NAMES: {
        # DN of corresponding APIC VRF; can be specified only on create.
        # Change 'allow_put' if updates on other DNs is allowed later,
        # and validate that VRF DN may not be updated.
        'allow_post': True, 'allow_put': False,
        'is_visible': True,
        'default': None,
        'validate': {
            'type:dict_or_none': {
                VRF: {'type:string': None,
                      'required': True}
            }
        }
    }
}


EXTENDED_ATTRIBUTES_2_0 = {
    port_def.COLLECTION_NAME: dict(
        list(APIC_ATTRIBUTES.items()) + list(PORT_ATTRIBUTES.items())),
    net_def.COLLECTION_NAME: dict(
        list(APIC_ATTRIBUTES.items()) + list(EXT_NET_ATTRIBUTES.items()) +
        list(NET_ATTRIBUTES.items())),
    subnet_def.COLLECTION_NAME: dict(
        list(APIC_ATTRIBUTES.items()) + list(EXT_SUBNET_ATTRIBUTES.items())),
    as_def.COLLECTION_NAME: dict(
        list(APIC_ATTRIBUTES.items()) + list(ADDRESS_SCOPE_ATTRIBUTES.items()))
}


class Cisco_apic(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Cisco APIC"

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return ("Extension exposing mapping of Neutron resources to Cisco "
                "APIC constructs")

    @classmethod
    def get_updated(cls):
        return "2016-03-31T12:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib import context as ctx
from neutron_lib import rpc as n_rpc
from oslo_config import cfg
import unittest2

from gbpservice.nfp.proxy_agent.notifications import pull

pull_notification = pull.PullNotification


class TestContext(object):

    def get_context_dict(self):
        try:
            context = ctx.Context(user_id='some_user',
                    tenant_id='some_tenant', is_advsvc=True)
        except Exception:
            context = ctx.Context(user_id='admin',
                    tenant_id='admin', is_advsvc=True, is_admin=True)
        return context.to_dict()


"""Common class for pull notification tests"""


class PullNotificationTestCase(unittest2.TestCase):

    def setUp(self):
        n_rpc.init(cfg.CONF)
        self.p_notification = pull_notification('sc', 'conf')
        self.context = TestContext().get_context_dict()
        self.ev = ''
        self.import_lib = 'gbpservice.nfp.lib.transport'
        self.import_cast = 'oslo_messaging.rpc.client._CallContext.cast'

    def _resp_base_structure(self, requester):
        response_data = [{
            'info': {
                'context': {
                    'neutron_context': self.context,
                    'requester': requester}
            }}]
        return response_data

    def _cast(self, context, method, **kwargs):
        return

    def _resp_data_nso(self, conf):
        response_data = self._resp_base_structure('service_orch')
        return response_data

    def _resp_data_ndo(self, conf):
        response_data = self._resp_base_structure('device_orch')
        return response_data

    def _resp_data_nco(self, conf):
        response_data = self._resp_base_structure('nas_service')
        return response_data

    def test_nco_pull_notifications(self):
        import_get = self.import_lib + '.get_response_from_configurator'
        with mock.patch(import_get) as (
            mock_get), mock.patch(self.import_cast) as (
            mock_cast):
            mock_get.side_effect = self._resp_data_nco
            mock_cast.side_effect = self._cast
            self.p_notification.pull_notifications(self.ev)

    def test_nso_pull_notifications(self):
        import_get = self.import_lib + '.get_response_from_configurator'
        with mock.patch(import_get) as (
            mock_get), mock.patch(self.import_cast) as (
            mock_cast):
            mock_get.side_effect = self._resp_data_nso
            mock_cast.side_effect = self._cast
            self.p_notification.pull_notifications(self.ev)

    def test_ndo_pull_notifications(self):
        import_get = self.import_lib + '.get_response_from_configurator'
        with mock.patch(import_get) as (
            mock_get), mock.patch(self.import_cast) as (
            mock_cast):
            mock_get.side_effect = self._resp_data_ndo
            mock_cast.side_effect = self._cast
            self.p_notification.pull_notifications(self.ev)



# Copyright (c) 2020 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# REVISIT: Eliminate this module as soon as definitions from
# neutron.db.api, which is eliminated in stein, are no longer
# needed. Please DO NOT add any definition to this module that is not
# a direct alias of a definion in the version of neutron_lib.db.api
# corresponding to the newest neutron branch supported by this
# repository.

import sqlalchemy

import neutron.objects.base as n_base
from neutron_lib.db import api
from neutron_lib.db import model_query
from neutron_lib.db import utils as db_utils

get_by_id = model_query.get_by_id
get_collection = model_query.get_collection
get_collection_count = model_query.get_collection_count
get_collection_query = model_query.get_collection_query
get_context_manager = api.get_context_manager
get_marker_obj = db_utils.get_marker_obj
get_reader_session = api.get_reader_session
get_writer_session = api.get_writer_session
is_retriable = api.is_retriable
resource_fields = db_utils.resource_fields
retry_db_errors = api.retry_db_errors
retry_if_session_inactive = api.retry_if_session_inactive
CONTEXT_READER = api.CONTEXT_READER
CONTEXT_WRITER = api.CONTEXT_WRITER


def get_session_from_obj(db_obj):
    # Support OVOs
    if isinstance(db_obj, n_base.NeutronObject):
        return db_obj.obj_context.session
    try:
        instance = sqlalchemy.inspect(db_obj)
        return instance.session
    except Exception:
        return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gbpservice.network.neutronv2 import client


class API(object):
    """API for interacting with the neutron 2.x API."""

    def _create_resource(self, context, resource, attrs):
        action = 'create_' + resource
        neutron = client.get_client(context)
        obj_creator = getattr(neutron, action)
        return obj_creator(attrs)[resource]

    def _show_resource(self, context, resource, resource_id):
        action = 'show_' + resource
        neutron = client.get_client(context)
        obj_method = getattr(neutron, action)
        return obj_method(resource_id)[resource]

    def _list_resources(self, context, resource, filters=None):
        filters = filters or {}
        resources = resource + 's'
        action = 'list_' + resources
        neutron = client.get_client(context)
        obj_lister = getattr(neutron, action)
        return obj_lister(**filters)[resources]

    def _update_resource(self, context, resource, resource_id, attrs):
        action = 'update_' + resource
        neutron = client.get_client(context)
        obj_updater = getattr(neutron, action)
        return obj_updater(resource_id, attrs)[resource]

    def _delete_resource(self, context, resource, resource_id):
        action = 'delete_' + resource
        neutron = client.get_client(context)
        obj_deleter = getattr(neutron, action)
        obj_deleter(resource_id)

    def create_network(self, context, network):
        return self._create_resource(context, 'network', network)

    def show_network(self, context, net_id):
        return self._show_resource(context, 'network', net_id)

    def list_networks(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'network', filters)

    def update_network(self, context, net_id, network):
        return self._update_resource(context, 'network', net_id, network)

    def delete_network(self, context, net_id):
        self._delete_resource(context, 'network', net_id)

    def create_subnet(self, context, subnet):
        return self._create_resource(context, 'subnet', subnet)

    def show_subnet(self, context, subnet_id):
        return self._show_resource(context, 'subnet', subnet_id)

    def list_subnets(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'subnet', filters)

    def update_subnet(self, context, subnet_id, subnet):
        return self._update_resource(context, 'subnet', subnet_id, subnet)

    def delete_subnet(self, context, subnet_id):
        self._delete_resource(context, 'subnet', subnet_id)

    def create_port(self, context, port):
        return self._create_resource(context, 'port', port)

    def show_port(self, context, port_id):
        return self._show_resource(context, 'port', port_id)

    def list_ports(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'port', filters)

    def update_port(self, context, port_id, port):
        return self._update_resource(context, 'port', port_id, port)

    def delete_port(self, context, port_id):
        self._delete_resource(context, 'port', port_id)

    def create_security_group(self, context, sg):
        return self._create_resource(context, 'security_group', sg)

    def show_security_group(self, context, sg_id):
        return self._show_resource(context, 'security_group', sg_id)

    def list_security_groups(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'security_group', filters)

    def update_security_group(self, context, sg_id, sg):
        return self._update_resource(context, 'security_group', sg_id, sg)

    def delete_security_group(self, context, sg_id):
        self._delete_resource(context, 'security_group', sg_id)

    def create_security_group_rule(self, context, rule):
        return self._create_resource(context, 'security_group_rule', rule)

    def show_security_group_rule(self, context, rule_id):
        return self._show_resource(context, 'security_group_rule', rule_id)

    def list_security_group_rules(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'security_group_rule', filters)

    # REVISIT(yi): update_security_group_rule not supported in neutron yet
    # def update_security_group_rule(self, context, rule_id, rule):
    #     return self._update_resource(context,
    #                                  'security_group_rule',
    #                                  rule_id,
    #                                  rule)

    def delete_security_group_rule(self, context, rule_id):
        self._delete_resource(context, 'security_group_rule', rule_id)

    def create_router(self, context, router):
        return self._create_resource(context, 'router', router)

    def show_router(self, context, router_id):
        return self._show_resource(context, 'router', router_id)

    def list_routers(self, context, filters=None):
        filters = filters or {}
        return self._list_resources(context, 'router', filters)

    def update_router(self, context, router_id, router):
        return self._update_resource(context, 'router', router_id, router)

    def delete_router(self, context, router_id):
        self._delete_resource(context, 'router', router_id)

    def add_router_interface(self, context, router_id, interface):
        return client.get_client(context).add_interface_router(router_id,
                                                               interface)

    def remove_router_interface(self, context, router_id, interface):
        return client.get_client(context).remove_interface_router(router_id,
                                                                  interface)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters as conv
from neutron_lib.api import extensions
from neutron_lib import constants as nlib_const
from neutron_lib import exceptions as nexc
from oslo_config import cfg

from gbpservice._i18n import _
from gbpservice.neutron.extensions import group_policy as gp
from gbpservice.neutron.services.grouppolicy.common import exceptions as gp_exc

PROXY_TYPE_L2 = 'l2'
PROXY_TYPE_L3 = 'l3'
DEFAULT_PROXY_TYPE = PROXY_TYPE_L3
PROXY_GROUP = 'proxy_group'

opts = [
    cfg.StrOpt('default_proxy_ip_pool',
               default='192.168.0.0/16',
               help=_("Proxy IP pool for implicitly created default "
                      "L3 policies, from which subnets are allocated for "
                      "policy target groups with proxy_group_id set to a "
                      "valid value.")),
    cfg.IntOpt('default_proxy_subnet_prefix_length',
               default=28,
               help=_("Proxy Subnet prefix length for implicitly created "
                      "default L3 polices, controlling size of subnets "
                      "allocated for policy target groups with proxy_group_id "
                      "set to a valid value.")),
]

cfg.CONF.register_opts(opts, "group_policy_proxy_group")
PROXY_CONF = cfg.CONF.group_policy_proxy_group


class ProxyGroupBadRequest(gp_exc.GroupPolicyBadRequest):
    message = _("Invalid input for Proxy Group extension, reason: %(msg)s")


class InvalidProxiedGroup(nexc.InvalidInput, ProxyGroupBadRequest):
    message = _("Proxied group %(group_id)s already has a proxy.")


class ProxyTypeSetWithoutProxiedPTG(nexc.InvalidInput, ProxyGroupBadRequest):
    message = _("Proxy type can't be set without a proxied PTG.")


class InvalidProxyGatewayGroup(nexc.InvalidInput, ProxyGroupBadRequest):
    message = _("Proxy gateway can't be set for non proxy PTG %(group_id)s.")


EXTENDED_ATTRIBUTES_2_0 = {
    gp.POLICY_TARGET_GROUPS: {
        'proxied_group_id': {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:uuid_or_none': None}, 'is_visible': True,
            'default': nlib_const.ATTR_NOT_SPECIFIED,
            'enforce_policy': True},
        'proxy_type': {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:values': ['l2', 'l3', None]},
            'is_visible': True, 'default': nlib_const.ATTR_NOT_SPECIFIED,
            'enforce_policy': True},
        'proxy_group_id': {
            'allow_post': False, 'allow_put': False,
            'validate': {'type:uuid_or_none': None}, 'is_visible': True,
            'enforce_policy': True},
        'enforce_service_chains': {
            'allow_post': True, 'allow_put': False, 'default': True,
            'convert_to': conv.convert_to_boolean,
            'is_visible': True, 'required_by_policy': True,
            'enforce_policy': True},
        # TODO(ivar): The APIs should allow the creation of a group with a
        # custom subnet prefix length. It may be useful for both the proxy
        # groups and traditional ones.
    },
    gp.L3_POLICIES: {
        'proxy_ip_pool': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:string_or_none': None},
                          'default': PROXY_CONF.default_proxy_ip_pool,
                          'is_visible': True},
        'proxy_subnet_prefix_length': {
            'allow_post': True, 'allow_put': True,
            'convert_to': conv.convert_to_int,
            'default': conv.convert_to_int(
                PROXY_CONF.default_proxy_subnet_prefix_length),
            'is_visible': True},
        # Proxy IP version is the same as the standard L3 pool ip version
    },
    gp.POLICY_TARGETS: {
        # This policy target will be used to reach the -proxied- PTG
        'proxy_gateway': {
            'allow_post': True, 'allow_put': False, 'default': False,
            'convert_to': conv.convert_to_boolean,
            'is_visible': True, 'required_by_policy': True,
            'enforce_policy': True},
        # This policy target is the default gateway for the -current- PTG
        # Only for internal use.
        'group_default_gateway': {
            'allow_post': True, 'allow_put': False, 'default': False,
            'convert_to': conv.convert_to_boolean,
            'is_visible': True, 'required_by_policy': True,
            'enforce_policy': True},
    },
}


class Driver_proxy_group(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "group policy poxy group extension"

    @classmethod
    def get_alias(cls):
        return PROXY_GROUP

    @classmethod
    def get_description(cls):
        return _("Add proxy_group_id attribute to policy target groups.")

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/ext/neutron/grouppolicy/"
                "proxy_group/api/v1.0")

    @classmethod
    def get_updated(cls):
        return "2015-08-03T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import re

from neutron.api import extensions as neutron_extensions
from neutron.api.v2 import resource_helper
from neutron_lib.api import converters as conv
from neutron_lib.api import extensions
from neutron_lib.api import validators as valid
from neutron_lib import constants as nlib_const
from neutron_lib import exceptions as nexc
from neutron_lib.plugins import constants
from neutron_lib.services import base as service_base
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import six

from gbpservice._i18n import _
from gbpservice.neutron import extensions as gbp_extensions
from gbpservice.neutron.extensions import patch  # noqa
from gbpservice.neutron.services.grouppolicy.common import (
    constants as gp_constants)


# The code below is a monkey patch of key Neutron's modules. This is needed for
# the GBP service to be loaded correctly. GBP extensions' path is added
# to Neutron's so that it's found at extension scanning time.
neutron_extensions.append_api_extensions_path(gbp_extensions.__path__)

LOG = logging.getLogger(__name__)

opts = [
    cfg.StrOpt('default_ip_pool',
               default='10.0.0.0/8',
               help=_("IP pool for implicitly created default L3 policies, "
                      "from which subnets are allocated for policy target "
                      "groups.")),
]

cfg.CONF.register_opts(opts, "group_policy_group")
GBP_CONF = cfg.CONF.group_policy_group


# Group Policy Exceptions
class GbpResourceNotFound(nexc.NotFound):
    message = _("Group Policy resource %(identity)s with id %(id)s could not "
                "be found")


class PolicyTargetNotFound(nexc.NotFound):
    message = _("Policy Target %(policy_target_id)s could not be found")


class PolicyTargetGroupNotFound(nexc.NotFound):
    message = _("Policy Target Group %(policy_target_group_id)s could not "
                "be found")


class ApplicationPolicyGroupNotFound(nexc.NotFound):
    message = _("Application Policy Group %(application_policy_group_id)s "
                "could not be found")


class ManagementPolicyTargetGroupExists(nexc.BadRequest):
    message = _("Service Management Policy Target Group already exists for "
                "this tenant.")


class L2PolicyNotFound(nexc.NotFound):
    message = _("L2Policy %(l2_policy_id)s could not be found")


class L2PolicyInUse(nexc.InUse):
    message = _("Unable to complete operation, L2Policy %(l2_policy_id)s is "
                "in use")


class L3PolicyNotFound(nexc.NotFound):
    message = _("L3Policy %(l3_policy_id)s could not be found")


class L3PolicyInUse(nexc.InUse):
    message = _("Unable to complete operation, L3Policy %(l3_policy_id)s is "
                "in use")


class NetworkServicePolicyInUse(nexc.InUse):
    message = _("Unable to complete operation, NetworkServicePolicy "
                "%(network_service_policy_id)s is in use")


class NetworkServicePolicyNotFound(nexc.NotFound):
    message = _("NetworkServicePolicy %(network_service_policy_id)s "
                "could not be found")


class InvalidDefaultSubnetPrefixLength(nexc.InvalidInput):
    message = _("Default subnet prefix length %(length)s is invalid for "
                "ipv%(protocol)s")


class SubnetPrefixLengthExceedsIpPool(nexc.InvalidInput):
    message = _("IP pool %(ip_pool)s prefix greater "
                "than subnet mask %(subnet_size)s")


class InvalidIpPoolSize(nexc.InvalidInput):
    message = _("IP pool %(ip_pool)s is invalid:%(err_msg)s"
                "Pool size=%(size)s")


class InvalidIpPoolPrefixLength(nexc.InvalidInput):
    message = _("IP pool %(ip_pool)s is invalid:%(err_msg)s"
                "Prefix Length=%(prefixlen)s")


class InvalidIpPoolVersion(nexc.InvalidInput):
    message = _("%(ip_pool)s is not a ipv%(version)s address.")


class PolicyClassifierNotFound(nexc.NotFound):
    message = _("PolicyClassifier %(policy_classifier_id)s could not be found")


class PolicyClassifierInUse(nexc.InUse):
    message = _("Unable to complete operation, PolicyClassifier "
                "%(policy_classifier_id)s is in use")


class PolicyActionNotFound(nexc.NotFound):
    message = _("PolicyAction %(policy_action_id)s could not be found")


class PolicyActionInUse(nexc.InUse):
    message = _("Unable to complete operation, PolicyAction "
                "%(policy_action_id)s is in use")


class PolicyRuleNotFound(nexc.NotFound):
    message = _("PolicyRule %(policy_rule_id)s could not be found")


class PolicyRuleInUse(nexc.InUse):
    message = _("Unable to complete operation, PolicyRule %(policy_rule_id)s "
                "is in use")


class PolicyRuleSetNotFound(nexc.NotFound):
    message = _("Policy Rule Set %(policy_rule_set_id)s could not be found")


class PolicyRuleSetInUse(nexc.InUse):
    message = _("Unable to complete operation, PolicyRuleSet "
                "%(policy_rule_set_id)s is in use")


class ExternalPolicyNotFound(nexc.NotFound):
    message = _("External Policy %(id)s could not be found")


class ExternalSegmentNotFound(nexc.NotFound):
    message = _("External Segment %(id)s could not be found")


class ExternalSegmentInUse(nexc.InUse):
    message = _("Unable to complete operation, External Segment "
                "%(es_id)s is in use")


class NATPoolNotFound(nexc.NotFound):
    message = _("NAT Pool %(id)s could not be found")


class BadPolicyRuleSetRelationship(nexc.BadRequest):
    message = _("Policy Rule Set %(parent_id)s is an invalid parent for "
                "%(child_id)s, make sure that child policy_rule_set has no "
                "children, or that you are not creating a relationship loop")


class ThreeLevelPolicyRuleSetHierarchyNotSupported(nexc.BadRequest):
    message = _("Can't add children to policy_rule_set %(policy_rule_set_id)s "
                "which already has a parent. Only one level "
                "of policy_rule_set hierarchy supported.")


class GroupPolicyInvalidPortValue(nexc.InvalidInput):
    message = _("Invalid value for port %(port)s")


class GroupPolicyInvalidProtocol(nexc.InvalidInput):
    message = _("Protocol %(protocol)s is not supported. "
                "Only protocol values %(values)s and their integer "
                "representation (0 to 255) are supported.")


class IpAddressOverlappingInExternalSegment(nexc.BadRequest):
    message = _("One or more requested IP addresses are already allocated for "
                "External Segment %(es_id)s.")


# Group Policy Values
gp_supported_actions = [None, gp_constants.GP_ACTION_ALLOW,
                        gp_constants.GP_ACTION_REDIRECT]
gp_supported_directions = [None, gp_constants.GP_DIRECTION_IN,
                           gp_constants.GP_DIRECTION_OUT,
                           gp_constants.GP_DIRECTION_BI]
gp_supported_protocols = [None, nlib_const.PROTO_NAME_TCP,
                          nlib_const.PROTO_NAME_UDP,
                          nlib_const.PROTO_NAME_ICMP]
gp_network_service_param_types = [
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE_IP_SINGLE,
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE_IP_POOL,
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE_QOS_MAX,    # expects integer value
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE_QOS_BURST,  # expects integer value
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE_STRING]
gp_network_service_param_keys = [
    gp_constants.GP_NETWORK_SVC_PARAM_TYPE,
    gp_constants.GP_NETWORK_SVC_PARAM_NAME,
    gp_constants.GP_NETWORK_SVC_PARAM_VALUE]
gp_network_service_param_values = [
    gp_constants.GP_NETWORK_SVC_PARAM_VALUE_SELF_SUBNET,
    gp_constants.GP_NETWORK_SVC_PARAM_VALUE_NAT_POOL]


# Group Policy input value conversion and validation functions
def convert_protocol(value):
    if value is None:
        return
    try:
        val = int(value)
        if val >= 0 and val <= 255:
            return value
        raise GroupPolicyInvalidProtocol(
            protocol=value, values=gp_supported_protocols)
    except (ValueError, TypeError):
        protocol = value.lower()
        if protocol in gp_supported_protocols:
            return protocol
        raise GroupPolicyInvalidProtocol(
            protocol=value, values=gp_supported_protocols)
    except AttributeError:
        raise GroupPolicyInvalidProtocol(
            protocol=value, values=gp_supported_protocols)


def convert_action_to_case_insensitive(value):
    if value is None:
        return
    else:
        return value.lower()


def convert_port_to_string(value):
    if value is None:
        return
    else:
        return str(value)


def convert_to_int_if_needed(value):
    if not value or value is nlib_const.ATTR_NOT_SPECIFIED:
        return value
    else:
        return conv.convert_to_int(value)


def _validate_gbp_port_range(data, key_specs=None):
    if data is None:
        return
    data = str(data)
    ports = data.split(':', 1)
    lower_range = 0

    for p in ports:
        try:
            val = int(p)
            if val <= 0 or val > 65535:
                msg = ("Invalid port '%s', valid range 0 < port < 65536") % p
                LOG.debug(msg)
                return msg
            if val <= lower_range:
                msg_dict = dict(p1=lower_range, p2=val)
                msg = ("Invalid port range: %(p1)s:%(p2)s, "
                       "valid range 0 < port1 < port2") % msg_dict
                LOG.debug(msg)
                return msg
            lower_range = val
        except (ValueError, TypeError):
            msg = ("Port value '%s' is not a valid number") % p
            LOG.debug(msg)
            return msg


def _validate_network_svc_params(data, key_specs=None):
    if data is None:
        return
    # A valid network_svc_params dict is:
    # [{'type': <type>, 'name': <param_name>, 'value': <param_value>}]
    # e.g. [{'type': 'ip_single', 'name': 'vip', 'value': 'self_subnet'}]
    # The type and value are validated, the name is treated as a literal.
    # The name of the param is chosen by the service chain implementation,
    # and as such is validated by the service chain provider.
    # The supported types are defined in gp_network_service_param_types.
    # The supported values are defined in gp_network_service_param_values,
    # but the values are not validated when the tpye is 'string'.
    if not isinstance(data, list):
        msg = ("'%s' is not a list") % data
        LOG.debug(msg)
        return msg
    for d in data:
        if not isinstance(d, dict):
            msg = ("'%s' is not a dictionary") % d
            LOG.debug(msg)
            return msg
        if set(d) != set(gp_network_service_param_keys):
            s = ", ".join(set(d) - set(gp_network_service_param_keys))
            msg = ("Unknown key(s) '%s' in network service params") % s
            LOG.debug(msg)
            return msg
        if d['type'] not in gp_network_service_param_types:
            msg = ("Network service param type(s) '%s' not supported") % (
                d['type'])
            LOG.debug(msg)
            return msg
        if d['type'] not in (gp_constants.GP_NETWORK_SVC_PARAM_TYPE_STRING,
                             gp_constants.GP_NETWORK_SVC_PARAM_TYPE_QOS_MAX,
                             gp_constants.GP_NETWORK_SVC_PARAM_TYPE_QOS_BURST):
            if d['value'] not in gp_network_service_param_values:
                msg = ("Network service param value '%s' is not "
                       "supported") % d['value']
                LOG.debug(msg)
                return msg


def _validate_external_dict(data, key_specs=None):
    if data is None:
        return
    if not isinstance(data, dict):
        msg = ("'%s' is not a dictionary") % data
        LOG.debug(msg)
        return msg
    for d in data:
        if not uuidutils.is_uuid_like(d):
            msg = ("'%s' is not a valid UUID") % d
            LOG.debug(msg)
            return msg
        if not isinstance(data[d], list):
            msg = ("'%s' is not a list") % data[d]
            LOG.debug(msg)
            return msg


def _validate_gbproutes(data, valid_values=None):
    # Shamelessly copied from Neutron, will pass even if nexthop is valid
    if not isinstance(data, list):
        msg = ("Invalid data format for hostroute: '%s'") % data
        LOG.debug(msg)
        return msg

    expected_keys = ['destination', 'nexthop']
    hostroutes = []
    for hostroute in data:
        msg = valid._verify_dict_keys(expected_keys, hostroute)
        if msg:
            LOG.debug(msg)
            return msg
        msg = valid.validate_subnet(hostroute['destination'])
        if msg:
            LOG.debug(msg)
            return msg
        if hostroute['nexthop']:
            msg = valid.validate_ip_address(hostroute['nexthop'])
        if msg:
            LOG.debug(msg)
            return msg
        if hostroute in hostroutes:
            msg = ("Duplicate hostroute '%s'") % hostroute
            LOG.debug(msg)
            return msg
        hostroutes.append(hostroute)


def _validate_gbp_resource_name(data, valid_values=None):
    # Any REST API defined GBP resource name is restricted to 128 characters
    return valid.validate_string(data, max_len=128)


AUTO_PTG_REGEX = re.compile('auto[0-9a-f]{32}\Z', re.I)


def _validate_gbp_uuid_or_none(data, valid_values=None):
    if data is not None:
        if not bool(AUTO_PTG_REGEX.match(data)):
            return valid.validate_uuid_or_none(data)


valid.validators['type:gbp_port_range'] = _validate_gbp_port_range
valid.validators['type:network_service_params'] = _validate_network_svc_params
valid.validators['type:external_dict'] = _validate_external_dict
valid.validators['type:gbproutes'] = _validate_gbproutes
valid.validators['type:gbp_resource_name'] = _validate_gbp_resource_name
valid.validators['type:gbp_uuid_or_none'] = _validate_gbp_uuid_or_none


POLICY_TARGETS = 'policy_targets'
POLICY_TARGET_GROUPS = 'policy_target_groups'
L2_POLICIES = 'l2_policies'
L3_POLICIES = 'l3_policies'
POLICY_CLASSIFIERS = 'policy_classifiers'
POLICY_ACTIONS = 'policy_actions'
POLICY_RULES = 'policy_rules'
POLICY_RULE_SETS = 'policy_rule_sets'
NETWORK_SERVICE_POLICIES = 'network_service_policies'
EXTERNAL_POLICIES = 'external_policies'
EXTERNAL_SEGMENTS = 'external_segments'
NAT_POOLS = 'nat_pools'
APPLICATION_POLICY_GROUPS = 'application_policy_groups'


RESOURCE_ATTRIBUTE_MAP = {
    POLICY_TARGETS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None}, 'default': '',
                 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'policy_target_group_id': {'allow_post': True, 'allow_put': True,
                                   'validate': {'type:gbp_uuid_or_none': None},
                                   'required': True, 'is_visible': True},
        'cluster_id': {'allow_post': True, 'allow_put': True,
                       'validate': {'type:string': None},
                       'default': '', 'is_visible': True}
    },
    POLICY_TARGET_GROUPS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'policy_targets': {'allow_post': False, 'allow_put': False,
                           'validate': {'type:uuid_list': None},
                           'convert_to': conv.convert_none_to_empty_list,
                           'default': None, 'is_visible': True},
        'l2_policy_id': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:uuid_or_none': None},
                         'default': None, 'is_visible': True},
        'application_policy_group_id': {'allow_post': True, 'allow_put': True,
                                        'validate':
                                        {'type:uuid_or_none': None},
                                        'default': None, 'is_visible': True},
        'provided_policy_rule_sets': {'allow_post': True, 'allow_put': True,
                                      'validate': {'type:dict_or_none': None},
                                      'convert_to':
                                      conv.convert_none_to_empty_dict,
                                      'default': None, 'is_visible': True},
        'consumed_policy_rule_sets': {'allow_post': True, 'allow_put': True,
                                      'validate': {'type:dict_or_none': None},
                                      'convert_to':
                                      conv.convert_none_to_empty_dict,
                                      'default': None, 'is_visible': True},
        'network_service_policy_id': {'allow_post': True, 'allow_put': True,
                                      'validate': {'type:uuid_or_none': None},
                                      'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                            'default': False,
                            'convert_to': conv.convert_to_boolean,
                            'is_visible': True, 'required_by_policy': True,
                            'enforce_policy': True},
        'service_management': {'allow_post': True, 'allow_put': True,
                               'default': False,
                               'convert_to': conv.convert_to_boolean,
                               'is_visible': True, 'required_by_policy': True,
                               'enforce_policy': True},
    },
    APPLICATION_POLICY_GROUPS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'policy_target_groups': {'allow_post': False, 'allow_put': False,
                                 'validate': {'type:uuid_list': None},
                                 'convert_to': conv.convert_none_to_empty_list,
                                 'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                            'default': False,
                            'convert_to': conv.convert_to_boolean,
                            'is_visible': True, 'required_by_policy': True,
                            'enforce_policy': True},
    },
    L2_POLICIES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'policy_target_groups': {'allow_post': False, 'allow_put': False,
                                 'validate': {'type:uuid_list': None},
                                 'convert_to': conv.convert_none_to_empty_list,
                                 'default': None, 'is_visible': True},
        'l3_policy_id': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:uuid_or_none': None},
                         'default': None, 'is_visible': True,
                         'required': True},
        'inject_default_route': {'allow_post': True, 'allow_put': True,
                                 'default': True, 'is_visible': True,
                                 'convert_to': conv.convert_to_boolean,
                                 'required': False},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                            'default': False,
                            'convert_to': conv.convert_to_boolean,
                            'is_visible': True, 'required_by_policy': True,
                            'enforce_policy': True},
        # TODO(Sumit): uncomment when supported in data path
        # 'allow_broadcast': {'allow_post': True, 'allow_put': True,
        #                    'default': True, 'is_visible': True,
        #                    'convert_to': conv.convert_to_boolean,
        #                    'required': False},
    },
    L3_POLICIES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'ip_version': {'allow_post': True, 'allow_put': False,
                       'convert_to': conv.convert_to_int,
                       # The value 46 is used to indicate dual-stack
                       # (IPv4 and IPv6)
                       'validate': {'type:values': [4, 6, 46]},
                       'default': 4, 'is_visible': True},
        'ip_pool': {'allow_post': True, 'allow_put': False,
                    'validate': {'type:string_or_none': None},
                    'default': GBP_CONF.default_ip_pool, 'is_visible': True},
        'subnet_prefix_length': {'allow_post': True, 'allow_put': True,
                                 'convert_to': conv.convert_to_int,
                                 # This parameter only applies to ipv4
                                 # prefixes. For IPv4 legal values are
                                 # 2 to 30. For ipv6, this parameter
                                 # is ignored
                                 'default': 24, 'is_visible': True},
        'l2_policies': {'allow_post': False, 'allow_put': False,
                        'validate': {'type:uuid_list': None},
                        'convert_to': conv.convert_none_to_empty_list,
                        'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                            'default': False,
                            'convert_to': conv.convert_to_boolean,
                            'is_visible': True, 'required_by_policy': True,
                            'enforce_policy': True},
        'external_segments': {
            'allow_post': True, 'allow_put': True, 'default': None,
            'validate': {'type:external_dict': None},
            'convert_to': conv.convert_none_to_empty_dict, 'is_visible': True},
    },
    POLICY_CLASSIFIERS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True,
                      'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'protocol': {'allow_post': True, 'allow_put': True,
                     'is_visible': True, 'default': None,
                     'convert_to': convert_protocol},
        'port_range': {'allow_post': True, 'allow_put': True,
                       'validate': {'type:gbp_port_range': None},
                       'convert_to': convert_port_to_string,
                       'default': None, 'is_visible': True},
        'direction': {'allow_post': True, 'allow_put': True,
                      'validate': {'type:values': gp_supported_directions},
                      'default': gp_constants.GP_DIRECTION_BI,
                      'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                            'default': False,
                            'convert_to': conv.convert_to_boolean,
                            'is_visible': True, 'required_by_policy': True,
                            'enforce_policy': True},
    },
    POLICY_ACTIONS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True,
                      'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'action_type': {'allow_post': True, 'allow_put': False,
                        'convert_to': convert_action_to_case_insensitive,
                        'validate': {'type:values': gp_supported_actions},
                        'is_visible': True, 'default': 'allow'},
        'action_value': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:uuid_or_none': None},
                         'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    POLICY_RULES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'enabled': {'allow_post': True, 'allow_put': True,
                    'default': True, 'convert_to': conv.convert_to_boolean,
                    'is_visible': True},
        'policy_classifier_id': {'allow_post': True, 'allow_put': True,
                                 'validate': {'type:uuid': None},
                                 'is_visible': True, 'required': True},
        'policy_actions': {'allow_post': True, 'allow_put': True,
                           'default': None, 'is_visible': True,
                           'validate': {'type:uuid_list': None},
                           'convert_to': conv.convert_none_to_empty_list},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    POLICY_RULE_SETS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '',
                 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True,
                      'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'parent_id': {'allow_post': False, 'allow_put': False,
                      'validate': {'type:uuid': None},
                      'is_visible': True},
        'child_policy_rule_sets': {'allow_post': True, 'allow_put': True,
                                   'default': None, 'is_visible': True,
                                   'validate': {'type:uuid_list': None},
                                   'convert_to':
                                   conv.convert_none_to_empty_list},
        'policy_rules': {'allow_post': True, 'allow_put': True,
                         'default': None, 'validate': {'type:uuid_list': None},
                         'convert_to': conv.convert_none_to_empty_list,
                         'is_visible': True},
        'consuming_policy_target_groups': {
            'allow_post': False, 'allow_put': False, 'default': None,
            'is_visible': True},
        'consuming_external_policies': {
            'allow_post': False, 'allow_put': False, 'default': None,
            'is_visible': True},
        'providing_policy_target_groups': {
            'allow_post': False, 'allow_put': False, 'default': None,
            'is_visible': True},
        'providing_external_policies': {
            'allow_post': False, 'allow_put': False, 'default': None,
            'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    NETWORK_SERVICE_POLICIES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'policy_target_groups': {'allow_post': False, 'allow_put': False,
                                 'validate': {'type:uuid_list': None},
                                 'convert_to': conv.convert_none_to_empty_list,
                                 'default': None, 'is_visible': True},
        'network_service_params': {'allow_post': True, 'allow_put': False,
                                   'validate':
                                   {'type:network_service_params': None},
                                   'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    EXTERNAL_POLICIES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'external_segments': {
            'allow_post': True, 'allow_put': True, 'default': None,
            'validate': {'type:uuid_list': None},
            'convert_to': conv.convert_none_to_empty_list, 'is_visible': True},
        'provided_policy_rule_sets': {'allow_post': True, 'allow_put': True,
                                      'validate': {'type:dict_or_none': None},
                                      'convert_to':
                                      conv.convert_none_to_empty_dict,
                                      'default': None, 'is_visible': True},
        'consumed_policy_rule_sets': {'allow_post': True, 'allow_put': True,
                                      'validate': {'type:dict_or_none': None},
                                      'convert_to':
                                      conv.convert_none_to_empty_dict,
                                      'default': None, 'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    EXTERNAL_SEGMENTS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'ip_version': {'allow_post': True, 'allow_put': False,
                       'convert_to': conv.convert_to_int,
                       'validate': {'type:values': [4, 6]},
                       'default': 4, 'is_visible': True},
        'cidr': {'allow_post': True, 'allow_put': False,
                 'validate': {'type:subnet': None},
                 'default': '172.16.0.0/12', 'is_visible': True},
        'external_policies': {
            'allow_post': False, 'allow_put': False, 'default': None,
            'validate': {'type:uuid_list': None},
            'convert_to': conv.convert_none_to_empty_list, 'is_visible': True},
        'external_routes': {
            'allow_post': True, 'allow_put': True,
            'default': nlib_const.ATTR_NOT_SPECIFIED,
            'validate': {'type:gbproutes': None},
            'is_visible': True},
        'l3_policies': {'allow_post': False, 'allow_put': False,
                        'validate': {'type:uuid_list': None},
                        'convert_to': conv.convert_none_to_empty_list,
                        'default': None, 'is_visible': True},
        'port_address_translation': {
            'allow_post': True, 'allow_put': True,
            'default': False, 'convert_to': conv.convert_to_boolean,
            'is_visible': True, 'required_by_policy': True,
            'enforce_policy': True},
        'nat_pools': {
            'allow_post': False, 'allow_put': False,
            'validate': {'type:uuid_list': None},
            'default': [],
            'is_visible': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    NAT_POOLS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'ip_version': {'allow_post': True, 'allow_put': False,
                       'convert_to': conv.convert_to_int,
                       'validate': {'type:values': [4, 6]},
                       'default': 4, 'is_visible': True},
        'ip_pool': {'allow_post': True, 'allow_put': False,
                    'validate': {'type:subnet': None},
                    'is_visible': True},
        'external_segment_id': {'allow_post': True, 'allow_put': True,
                                'validate': {'type:uuid_or_none': None},
                                'is_visible': True, 'required': True},
        nlib_const.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    }
}


group_based_policy_quota_opts = [
    cfg.IntOpt('quota_l3_policy',
               default=-1,
               help=_('Number of L3 Policies allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_l2_policy',
               default=-1,
               help=_('Number of L2 Policies allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_application_policy_group',
               default=-1,
               help=_('Number of Application Policy Groups allowed per tenant.'
                      ' A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_target_group',
               default=-1,
               help=_('Number of Policy Target Groups allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_target',
               default=-1,
               help=_('Number of Policy Targets allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_classifier',
               default=-1,
               help=_('Number of Policy Classifiers allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_action',
               default=-1,
               help=_('Number of Policy Actions allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_rule',
               default=-1,
               help=_('Number of Policy Rules allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_policy_rule_set',
               default=-1,
               help=_('Number of Policy Rule Sets allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_external_policy',
               default=-1,
               help=_('Number of External Policies allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_external_segment',
               default=-1,
               help=_('Number of External Segments allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_nat_pool',
               default=-1,
               help=_('Number of NAT Pools allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_network_service_policy',
               default=-1,
               help=_('Number of Network Service Policies allowed per '
                      'tenant. A negative value means unlimited.')),
]
cfg.CONF.register_opts(group_based_policy_quota_opts, 'QUOTAS')


class Group_policy(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Group Policy Abstraction"

    @classmethod
    def get_alias(cls):
        return "group-policy"

    @classmethod
    def get_description(cls):
        return "Extension for Group Policy Abstraction"

    @classmethod
    def get_namespace(cls):
        return "https://wiki.openstack.org/wiki/Neutron/gp/v2.0/"

    @classmethod
    def get_updated(cls):
        return "2014-03-03T12:00:00-00:00"

    @classmethod
    def get_resources(cls):
        special_mappings = {
            'l2_policies': 'l2_policy', 'l3_policies': 'l3_policy',
            'network_service_policies': 'network_service_policy',
            'external_policies': 'external_policy'}
        plural_mappings = resource_helper.build_plural_mappings(
            special_mappings, RESOURCE_ATTRIBUTE_MAP)
        gbp_extensions.register_plurals(plural_mappings)
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   constants.GROUP_POLICY,
                                                   register_quota=True)

    @classmethod
    def get_plugin_interface(cls):
        return GroupPolicyPluginBase

    def update_attributes_map(self, attributes):
        super(Group_policy, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


@six.add_metaclass(abc.ABCMeta)
class GroupPolicyPluginBase(service_base.ServicePluginBase):

    def get_plugin_name(self):
        return constants.GROUP_POLICY

    def get_plugin_type(self):
        return constants.GROUP_POLICY

    def get_plugin_description(self):
        return 'Group Policy plugin'

    @abc.abstractmethod
    def get_policy_targets(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_target(self, context, policy_target_id, fields=None):
        pass

    @abc.abstractmethod
    def create_policy_target(self, context, policy_target):
        pass

    @abc.abstractmethod
    def update_policy_target(self, context, policy_target_id, policy_target):
        pass

    @abc.abstractmethod
    def delete_policy_target(self, context, policy_target_id):
        pass

    @abc.abstractmethod
    def get_policy_target_groups(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_target_group(self, context, policy_target_group_id,
                                fields=None):
        pass

    @abc.abstractmethod
    def create_policy_target_group(self, context, policy_target_group):
        pass

    @abc.abstractmethod
    def update_policy_target_group(self, context, policy_target_group_id,
                                   policy_target_group):
        pass

    @abc.abstractmethod
    def delete_policy_target_group(self, context, policy_target_group_id):
        pass

    @abc.abstractmethod
    def get_application_policy_groups(self, context, filters=None,
                                      fields=None):
        pass

    @abc.abstractmethod
    def get_application_policy_group(self, context,
                                     application_policy_group_id,
                                     fields=None):
        pass

    @abc.abstractmethod
    def create_application_policy_group(self, context,
                                        application_policy_group):
        pass

    @abc.abstractmethod
    def update_application_policy_group(self, context,
                                        application_policy_group_id,
                                        application_policy_group):
        pass

    @abc.abstractmethod
    def delete_application_policy_group(self, context,
                                        application_policy_group_id):
        pass

    @abc.abstractmethod
    def get_l2_policies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_l2_policy(self, context, l2_policy_id, fields=None):
        pass

    @abc.abstractmethod
    def create_l2_policy(self, context, l2_policy):
        pass

    @abc.abstractmethod
    def update_l2_policy(self, context, l2_policy_id, l2_policy):
        pass

    @abc.abstractmethod
    def delete_l2_policy(self, context, l2_policy_id):
        pass

    @abc.abstractmethod
    def get_l3_policies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_l3_policy(self, context, l3_policy_id, fields=None):
        pass

    @abc.abstractmethod
    def create_l3_policy(self, context, l3_policy):
        pass

    @abc.abstractmethod
    def update_l3_policy(self, context, l3_policy_id, l3_policy):
        pass

    @abc.abstractmethod
    def delete_l3_policy(self, context, l3_policy_id):
        pass

    @abc.abstractmethod
    def get_network_service_policies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_network_service_policy(
        self, context, network_service_policy_id, fields=None):
        pass

    @abc.abstractmethod
    def create_network_service_policy(self, context, network_service_policy):
        pass

    @abc.abstractmethod
    def update_network_service_policy(
        self, context, network_service_policy_id, network_service_policy):
        pass

    @abc.abstractmethod
    def delete_network_service_policy(
        self, context, network_service_policy_id):
        pass

    @abc.abstractmethod
    def get_policy_classifiers(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_classifier(self, context, policy_classifier_id,
                              fields=None):
        pass

    @abc.abstractmethod
    def create_policy_classifier(self, context, policy_classifier):
        pass

    @abc.abstractmethod
    def update_policy_classifier(self, context, policy_classifier_id,
                                 policy_classifier):
        pass

    @abc.abstractmethod
    def delete_policy_classifier(self, context, policy_classifier_id):
        pass

    @abc.abstractmethod
    def get_policy_actions(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_action(self, context, policy_action_id, fields=None):
        pass

    @abc.abstractmethod
    def create_policy_action(self, context, policy_action):
        pass

    @abc.abstractmethod
    def update_policy_action(self, context, policy_action_id, policy_action):
        pass

    @abc.abstractmethod
    def delete_policy_action(self, context, policy_action_id):
        pass

    @abc.abstractmethod
    def get_policy_rules(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_rule(self, context, policy_rule_id, fields=None):
        pass

    @abc.abstractmethod
    def create_policy_rule(self, context, policy_rule):
        pass

    @abc.abstractmethod
    def update_policy_rule(self, context, policy_rule_id, policy_rule):
        pass

    @abc.abstractmethod
    def delete_policy_rule(self, context, policy_rule_id):
        pass

    @abc.abstractmethod
    def create_policy_rule_set(self, context, policy_rule_set):
        pass

    @abc.abstractmethod
    def update_policy_rule_set(self, context, policy_rule_set_id,
                               policy_rule_set):
        pass

    @abc.abstractmethod
    def get_policy_rule_sets(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_policy_rule_set(self, context, policy_rule_set_id, fields=None):
        pass

    @abc.abstractmethod
    def delete_policy_rule_set(self, context, policy_rule_set_id):
        pass

    @abc.abstractmethod
    def create_external_policy(self, context, external_policy):
        pass

    @abc.abstractmethod
    def update_external_policy(self, context, external_policy_id,
                               external_policy):
        pass

    @abc.abstractmethod
    def get_external_policies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_external_policy(self, context, external_policy_id,
                            fields=None):
        pass

    @abc.abstractmethod
    def delete_external_policy(self, context, external_policy_id):
        pass

    @abc.abstractmethod
    def create_external_segment(self, context, external_segment):
        pass

    @abc.abstractmethod
    def update_external_segment(self, context, external_segment_id,
                                external_segment):
        pass

    @abc.abstractmethod
    def get_external_segments(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_external_segment(self, context, external_segment_id, fields=None):
        pass

    @abc.abstractmethod
    def delete_external_segment(self, context, external_segment_id):
        pass

    @abc.abstractmethod
    def create_nat_pool(self, context, nat_pool):
        pass

    @abc.abstractmethod
    def update_nat_pool(self, context, nat_pool_id, nat_pool):
        pass

    @abc.abstractmethod
    def get_nat_pools(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_nat_pool(self, context, nat_pool_id, fields=None):
        pass

    @abc.abstractmethod
    def delete_nat_pool(self, context, nat_pool_id):
        pass

    @abc.abstractmethod
    def start_rpc_listeners(self):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy.ext import baked

from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_auto_ptg_db as auto_ptg_db)
from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_intra_ptg_db as intra_ptg_db)
from gbpservice.neutron.db.grouppolicy import group_policy_db as gp_db
from gbpservice.neutron.extensions import cisco_apic_gbp
from gbpservice.neutron.extensions import group_policy as gpolicy
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

LOG = logging.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))


class AIMExtensionDriver(api.ExtensionDriver,
                         intra_ptg_db.ApicIntraPtgDBMixin,
                         auto_ptg_db.ApicAutoPtgDBMixin):
    _supported_extension_alias = cisco_apic_gbp.ALIAS
    _extension_dict = cisco_apic_gbp.EXTENDED_ATTRIBUTES_2_0

    def __init__(self):
        LOG.info("AIM Extension __init__")
        self._policy_driver = None

    @property
    def _pd(self):
        if not self._policy_driver:
            gbp_plugin = directory.get_plugin("GROUP_POLICY")
            policy_mgr = gbp_plugin.policy_driver_manager
            self._policy_driver = policy_mgr.policy_drivers['aim_mapping'].obj
        return self._policy_driver

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def _set_intra_ptg_allow(self, session, data, result):
        ptg = data['policy_target_group']

        query = BAKERY(lambda s: s.query(
            gp_db.PolicyTargetGroup))
        query += lambda q: q.filter_by(
            id=sa.bindparam('id'))
        ptg_db = query(session).params(
            id=result['id']).one()

        if not ptg_db:
            raise gpolicy.PolicyTargetGroupNotFound(
                policy_target_group_id=result['id'])
        if 'intra_ptg_allow' in ptg:
            self.set_intra_ptg_allow(
                session, policy_target_group_id=result['id'],
                intra_ptg_allow=ptg['intra_ptg_allow'])
            result['intra_ptg_allow'] = ptg['intra_ptg_allow']
        else:
            self._extend_ptg_dict_with_intra_ptg_allow(session, result)

    def _extend_ptg_dict_with_intra_ptg_allow(self, session, result):
        result['intra_ptg_allow'] = self.get_intra_ptg_allow(
            session, policy_target_group_id=result['id'])

    def process_create_policy_target_group(self, session, data, result):
        self._set_intra_ptg_allow(session, data, result)
        result['is_auto_ptg'] = bool(
            gpolicy.AUTO_PTG_REGEX.match(result['id']))
        self.set_is_auto_ptg(
            session, policy_target_group_id=result['id'],
            is_auto_ptg=result['is_auto_ptg'])

    def process_update_policy_target_group(self, session, data, result):
        self._set_intra_ptg_allow(session, data, result)

    def extend_policy_target_group_dict(self, session, result):
        self._extend_ptg_dict_with_intra_ptg_allow(session, result)
        result['is_auto_ptg'] = self.get_is_auto_ptg(
            session, policy_target_group_id=result['id'])
        self._pd.extend_policy_target_group_dict(session, result)

    def extend_application_policy_group_dict(self, session, result):
        self._pd.extend_application_policy_group_dict(session, result)

    def extend_policy_rule_dict(self, session, result):
        self._pd.extend_policy_rule_dict(session, result)

    def extend_policy_rule_set_dict(self, session, result):
        self._pd.extend_policy_rule_set_dict(session, result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add no nat cidrs network extension

Revision ID: e29a84f6a15f
Revises: 90e1d92a49b2

"""

# revision identifiers, used by Alembic.
revision = 'e29a84f6a15f'
down_revision = '90e1d92a49b2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_network_no_nat_cidrs',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('cidr', sa.String(64), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                        name='apic_aim_network_no_nat_cidrs_extn_fk_network',
                        ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'cidr')
    )


def downgrade():
    pass
